/*
 * kexec-elf-loongarch.c - kexec Elf loader for loongarch
 *
 * Copyright (C) 2021 Loongson Technology Co., Ltd.
 *
 * This source code is licensed under the GNU General Public License,
 * Version 2.  See the file COPYING for more details.
*/

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <elf.h>
#include <boot/elf_boot.h>
#include <ip_checksum.h>
#include "../../kexec.h"
#include "../../kexec-elf.h"
#include "../../kexec-syscall.h"
#include "crashdump-loongarch.h"
#include "kexec-loongarch.h"
#include <arch/options.h>


off_t initrd_base = 0;
off_t initrd_size = 0;

static const int probe_debug = 0;

#define BOOTLOADER         "kexec"
#define UPSZ(X) _ALIGN_UP(sizeof(X), 4)

#define CMDLINE_PREFIX "kexec "
static char cmdline_buf[COMMAND_LINE_SIZE] = CMDLINE_PREFIX;

/* Adds the rd_start= command line parameter to command line. */
static int cmdline_add_rd_start(char *cmdline, unsigned long addr)
{
	int cmdlen, len;
	char str[40], *ptr;

	ptr = str;
	strcpy(str, " rd_start=");
	ptr += strlen(str);
	ultoa(addr, ptr);
	len = strlen(str);
	cmdlen = strlen(cmdline) + len;
	if (cmdlen > (COMMAND_LINE_SIZE - 1))
		die("Command line overflow\n");
	strcat(cmdline, str);

	return 0;
}

/* Adds the rd_size= command line parameter to command line. */
static int cmdline_add_rd_size(char *cmdline, unsigned long addr)
{
	int cmdlen, len;
	char str[30], *ptr;

	ptr = str;
	strcpy(str, " rd_size=");
	ptr += strlen(str);
	ultoa(addr, ptr);
	len = strlen(str);
	cmdlen = strlen(cmdline) + len;
	if (cmdlen > (COMMAND_LINE_SIZE - 1))
		die("Command line overflow\n");
	strcat(cmdline, str);

	return 0;
}

int elf_loongarch_probe(const char *buf, off_t len)
{
	struct mem_ehdr ehdr;
	int result;
	result = build_elf_exec_info(buf, len, &ehdr, 0);
	if (result < 0) {
		if (probe_debug)
			fprintf(stderr, "Not an ELF executable.\n");
		goto out;
	}

	/* Verify the architecuture specific bits */
	if (ehdr.e_machine != EM_LOONGARCH) {
		/* for a different architecture */
		if (probe_debug) {
			fprintf(stderr, "Not LoongArch ELF executable.\n");
		}
		result = -1;
		goto out;
	}
	result = 0;
 out:
	free_elf_info(&ehdr);
	return result;
}

void elf_loongarch_usage(void)
{
	printf( "    --command-line=STRING Set the kernel command line to STRING\n"
		"    --append=STRING       Set the kernel command line to STRING\n"
		"    --reuse-cmdline       Use kernel command line from running system.\n"
		"    --initrd=FILE         Use FILE as initial ramdisk.\n"
	      );
}

int elf_loongarch_load(int argc, char **argv, const char *buf, off_t len,
	struct kexec_info *info)
{
	struct mem_ehdr ehdr;
	int command_line_len = 0;
	int result;
	size_t i;
	unsigned long cmdline_addr = 0;
	char *crash_cmdline;
	char *initrd_buf = NULL;
	unsigned long long kernel_addr = 0, kernel_size = 0;
	unsigned long pagesize = getpagesize();

	/*
	 * Need to append some command line parameters internally in case of
	 * taking crash dumps.
	 */
	if (info->kexec_flags & KEXEC_ON_CRASH) {
		crash_cmdline = xmalloc(COMMAND_LINE_SIZE);
		memset((void *)crash_cmdline, 0, COMMAND_LINE_SIZE);
	} else {
		crash_cmdline = NULL;
	}

	result = build_elf_exec_info(buf, len, &ehdr, 0);
	if (result < 0)
		die("ELF exec parse failed\n");

	/* Read in the PT_LOAD segments*/
	for (i = 0; i < ehdr.e_phnum; i++) {
		struct mem_phdr *phdr;
		phdr = &ehdr.e_phdr[i];
		if (phdr->p_type == PT_LOAD) {
			phdr->p_paddr = virt_to_phys(phdr->p_paddr);
			kernel_addr = phdr->p_paddr;
			kernel_size = phdr->p_memsz;
		}
	}

	/* Load the Elf data */
	result = elf_exec_load(&ehdr, info);
	if (result < 0)
		die("ELF exec load failed\n");

	info->entry = (void *)virt_to_phys(ehdr.e_entry);

	if (arch_options.command_line)
		command_line_len = strlen(arch_options.command_line) + 1;

	if (info->kexec_flags & KEXEC_ON_CRASH) {
		result = load_crashdump_segments(info, crash_cmdline, 0x0fffffff, 0);
		if (result < 0) {
			free(crash_cmdline);
			return -1;
		}
	}

	if (arch_options.command_line)
		strncat(cmdline_buf, arch_options.command_line, command_line_len);

	if (crash_cmdline) {
		strncat(cmdline_buf, crash_cmdline,
				sizeof(crash_cmdline) -
				strlen(crash_cmdline) - 1);
		free(crash_cmdline);
	}

	if (info->kexec_flags & KEXEC_ON_CRASH)
		/*
		 * In case of crashdump segment[0] is kernel.
		 * Put cmdline just after it.
		 */
		cmdline_addr = (unsigned long)info->segment[0].mem +
			info->segment[0].memsz;
	else
		cmdline_addr = 0x10000; /* Skip exception handlers */

	if (arch_options.initrd_file) {
		initrd_buf = slurp_decompress_file(arch_options.initrd_file, &initrd_size);

		initrd_base = add_buffer(info, initrd_buf, initrd_size,
					initrd_size, sizeof(void *),
					_ALIGN_UP(kernel_addr + kernel_size,
						pagesize), 0x0fffffff, 1);
		cmdline_add_rd_start(cmdline_buf, PAGE_OFFSET + initrd_base);
		cmdline_add_rd_size(cmdline_buf, initrd_size);
		dbgprintf("initrd_base: %lx, initrd_size: %lx\n", initrd_base, initrd_size);
	}

	/* This is a legacy method for command line passing used currently */
	add_buffer(info, cmdline_buf, sizeof(cmdline_buf),
			sizeof(cmdline_buf), sizeof(void *),
			cmdline_addr, 0x0fffffff, 1);
	dbgprintf("command line: %s\n", cmdline_buf);

	return 0;
}

